"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EvmUtils = void 0;
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const constants_1 = require("./constants");
const address_1 = require("./address");
const contracts_1 = require("./contracts");
const platform_1 = require("./platform");
// forces EvmUtils to implement PlatformUtils
var _ = EvmUtils;
/**
 * @category EVM
 */
// Provides runtime concrete value
var EvmUtils;
(function (EvmUtils) {
    function nativeTokenId(chain) {
        if (!isSupportedChain(chain))
            throw new Error(`invalid chain for EVM: ${chain}`);
        return {
            chain: chain,
            address: new address_1.EvmAddress(address_1.EvmZeroAddress), // TODO: fix weird type error
        };
    }
    EvmUtils.nativeTokenId = nativeTokenId;
    function isSupportedChain(chain) {
        const platform = (0, connect_sdk_1.chainToPlatform)(chain);
        return platform === platform_1.EvmPlatform.platform;
    }
    EvmUtils.isSupportedChain = isSupportedChain;
    function isNativeTokenId(chain, tokenId) {
        if (!isSupportedChain(chain))
            return false;
        if (tokenId.chain !== chain)
            return false;
        return tokenId.address.toString() === address_1.EvmZeroAddress;
    }
    EvmUtils.isNativeTokenId = isNativeTokenId;
    async function getDecimals(chain, rpc, token) {
        if (token === 'native')
            return (0, connect_sdk_1.nativeDecimals)(platform_1.EvmPlatform.platform);
        const tokenContract = contracts_1.EvmContracts.getTokenImplementation(rpc, new address_1.EvmAddress(token).toString());
        return tokenContract.decimals();
    }
    EvmUtils.getDecimals = getDecimals;
    async function getBalance(chain, rpc, walletAddr, token) {
        if (token === 'native')
            return rpc.getBalance(walletAddr);
        const tokenImpl = contracts_1.EvmContracts.getTokenImplementation(rpc, new address_1.EvmAddress(token).toString());
        return tokenImpl.balanceOf(walletAddr);
    }
    EvmUtils.getBalance = getBalance;
    async function getBalances(chain, rpc, walletAddr, tokens) {
        const balancesArr = await Promise.all(tokens.map(async (token) => {
            const balance = await getBalance(chain, rpc, walletAddr, token);
            const address = token === 'native' ? 'native' : new address_1.EvmAddress(token).toString();
            return { [address]: balance };
        }));
        return balancesArr.reduce((obj, item) => Object.assign(obj, item), {});
    }
    EvmUtils.getBalances = getBalances;
    async function sendWait(chain, rpc, stxns) {
        const txhashes = [];
        for (const stxn of stxns) {
            const txRes = await rpc.broadcastTransaction(stxn);
            txhashes.push(txRes.hash);
            if (chain === 'Celo') {
                console.error('TODO: override celo block fetching');
                continue;
            }
            // Wait for confirmation
            const txReceipt = await txRes.wait();
            if (txReceipt === null)
                continue; // TODO: throw error?
        }
        return txhashes;
    }
    EvmUtils.sendWait = sendWait;
    async function getCurrentBlock(rpc) {
        return await rpc.getBlockNumber();
    }
    EvmUtils.getCurrentBlock = getCurrentBlock;
    // Look up the Wormhole Canonical Network and Chain from the EVM chainId
    function chainFromChainId(eip155ChainId) {
        const ci = connect_sdk_1.encoding.bignum.decode(eip155ChainId);
        const networkChainPair = constants_1.evmChainIdToNetworkChainPair.get(ci);
        if (networkChainPair === undefined)
            throw new Error(`Unknown EVM chainId ${eip155ChainId}`);
        const [network, chain] = networkChainPair;
        return [network, chain];
    }
    EvmUtils.chainFromChainId = chainFromChainId;
    async function chainFromRpc(rpc) {
        const { chainId } = await rpc.getNetwork();
        return chainFromChainId(connect_sdk_1.encoding.bignum.encode(chainId, true));
    }
    EvmUtils.chainFromRpc = chainFromRpc;
})(EvmUtils || (exports.EvmUtils = EvmUtils = {}));
//# sourceMappingURL=platformUtils.js.map